@@ -1,4 +1,4 @@
-# pylint: disable=too-many-lines,too-many-instance-attributes,too-many-locals,unused-argument,no-self-use,assignment-from-none,invalid-name,ungrouped-imports
+# pylint: disable=too-many-lines,too-many-instance-attributes,too-many-locals,unused-argument,assignment-from-none,invalid-name,ungrouped-imports
 import sys
 import asyncio
 import csv
@@ -20,14 +20,17 @@ from typing import (
     TypeVar,
     Iterator,
     Coroutine,
+    Set,
+    Iterable,
 )
 from functools import wraps
 
-from google.api_core.exceptions import Forbidden
-from google.cloud import storage
+from cloudpathlib import AnyPath
+
+from sample_metadata.parser.cloudhelper import CloudHelper
+
 from sample_metadata.model.participant_upsert_body import ParticipantUpsertBody
 
-from sample_metadata.model_utils import async_wrap
 from sample_metadata.apis import SampleApi, SequenceApi, AnalysisApi, ParticipantApi
 from sample_metadata.models import (
     ParticipantUpsert,
@@ -67,7 +70,13 @@ ALL_EXTENSIONS = (
     + VCF_EXTENSIONS
 )
 
-rmatch = re.compile(r'[_\.-][Rr]\d')
+# construct rmatch string to capture all fastq patterns
+rmatch_str = (
+    r'[_\.-][Rr]?[12]('
+    + '|'.join(s.replace('.', '\\.') for s in FASTQ_EXTENSIONS)
+    + ')$'
+)
+rmatch = re.compile(rmatch_str)
 SingleRow = Dict[str, Any]
 GroupedRow = List[SingleRow]
 
@@ -78,6 +87,53 @@ SUPPORTED_READ_TYPES = Literal['fastq', 'bam', 'cram']
 SUPPORTED_VARIANT_TYPES = Literal['gvcf', 'vcf']
 
 
+class CustomDictReader(csv.DictReader):
+    """csv.DictReader that strips whitespace off headers"""
+
+    def __init__(
+        self,
+        *args,
+        key_map=None,
+        required_keys: Iterable[str] = None,
+        ignore_extra_keys=False,
+        **kwargs,
+    ):
+        super().__init__(*args, **kwargs)
+        self.key_map = key_map
+        self.ignore_extra_keys = ignore_extra_keys
+        self.required_keys = set(required_keys) if required_keys else None
+        self._custom_cached_fieldnames = None
+
+    @property
+    def fieldnames(self):
+        if not self._custom_cached_fieldnames:
+            fs = super().fieldnames
+            self._custom_cached_fieldnames = list(map(self.process_fieldname, fs))
+
+            if self.required_keys:
+                missing_keys = self.required_keys - set(self._custom_cached_fieldnames)
+                if missing_keys:
+                    raise ValueError('Missing keys: ' + ','.join(missing_keys))
+        return self._custom_cached_fieldnames
+
+    def process_fieldname(self, fieldname: str) -> str:
+        """
+        Process the fieldname
+        (default: strip leading / trailing whitespace)
+        """
+        if self.key_map:
+            for k, values in self.key_map.items():
+                if fieldname.lower() in values:
+                    return k
+
+            if not self.ignore_extra_keys:
+                raise ValueError(
+                    f'Key "{fieldname}" not found in provided key map: {", ".join(self.key_map.keys())}'
+                )
+
+        return fieldname.strip()
+
+
 class ParticipantMetaGroup:
     """Class for holding participant metadata grouped by id"""
 
@@ -120,12 +176,19 @@ class SequenceMetaGroup:
         self.meta = meta
 
 
-def chunk(iterable: Sequence[T], chunk_size=500) -> Iterator[Sequence[T]]:
+def chunk(iterable: Iterable[T], chunk_size=50) -> Iterator[List[T]]:
     """
     Chunk a sequence by yielding lists of `chunk_size`
     """
-    for i in range(0, len(iterable), chunk_size):
-        yield iterable[i : i + chunk_size]
+    chnk: List[T] = []
+    for element in iterable:
+        chnk.append(element)
+        if len(chnk) >= chunk_size:
+            yield chnk
+            chnk = []
+
+    if chnk:
+        yield chnk
 
 
 def run_as_sync(f):
@@ -148,19 +211,25 @@ def run_as_sync(f):
     return wrapper
 
 
-class GenericParser:  # pylint: disable=too-many-public-methods
+class GenericParser(
+    CloudHelper
+):  # pylint: disable=too-many-public-methods,too-many-arguments
     """Parser for VCGS manifest"""
 
-    def __init__(
+    def __init__(  # pylint: disable=too-many-arguments
         self,
         path_prefix: Optional[str],
-        sample_metadata_project: str,
+        search_paths: list[str],
+        project: str,
         default_sequence_type='genome',
         default_sequence_status='uploaded',
         default_sample_type='blood',
         default_analysis_type='qc',
         default_analysis_status='completed',
         skip_checking_gcs_objects=False,
+        key_map: Dict[str, str] = None,
+        ignore_extra_keys=False,
+        required_keys: Set[str] = None,
         verbose=True,
     ):
 
@@ -168,10 +237,14 @@ class GenericParser:  # pylint: disable=too-many-public-methods
         self.skip_checking_gcs_objects = skip_checking_gcs_objects
         self.verbose = verbose
 
-        if not sample_metadata_project:
+        self.key_map = key_map
+        self.required_keys = required_keys
+        self.ignore_extra_keys = ignore_extra_keys
+
+        if not project:
             raise ValueError('sample-metadata project is required')
 
-        self.sample_metadata_project = sample_metadata_project
+        self.project = project
 
         self.default_sequence_type: str = default_sequence_type
         self.default_sequence_status: str = default_sequence_status
@@ -186,24 +259,12 @@ class GenericParser:  # pylint: disable=too-many-public-methods
         self.bucket_clients: Dict[str, Any] = {}
 
         self.papi = ParticipantApi()
+        self.sapi = SampleApi()
         self.seqapi = SequenceApi()
 
-    @property
-    def client(self):
-        """Get GCP storage client"""
-        if not self._client:
-            self._client = storage.Client()
-        return self._client
+        super().__init__(search_paths)
 
-    def get_bucket(self, bucket_name):
-        """Get cached bucket client from optional bucket name"""
-        assert bucket_name
-        if bucket_name not in self.bucket_clients:
-            self.bucket_clients[bucket_name] = self.client.get_bucket(bucket_name)
-
-        return self.bucket_clients[bucket_name]
-
-    def file_path(self, filename: str) -> str:
+    def file_path(self, filename: str, raise_exception: bool = True) -> str | None:
         """
         Get complete filepath of filename:
         - Includes gs://{bucket} if relevant
@@ -213,94 +274,20 @@ class GenericParser:  # pylint: disable=too-many-public-methods
             return filename
 
         if not self.path_prefix:
-            raise FileNotFoundError(
-                f"Can't form full path to '{filename}' as no path_prefix was defined"
-            )
-
-        if self.client and not filename.startswith('/'):
-            assert self.default_bucket
-            return os.path.join(
-                'gs://',
-                self.default_bucket or '',
-                self.path_prefix or '',
-                filename or '',
-            )
+            fn = super().file_path(filename, raise_exception=raise_exception)
+            if not fn:
+                raise FileNotFoundError(
+                    f'Cannot form full path to "{filename}" as '
+                    'no path_prefix was defined'
+                )
+            return fn
 
         return os.path.join(self.path_prefix or '', filename)
 
-    @async_wrap
-    def get_blob(self, filename):
-        """Convenience function for getting blob from fully qualified GCS path"""
-        if not filename.startswith('gs://'):
-            raise ValueError('No blob available')
-
-        bucket_name, *components = filename[5:].split('/')
-        bucket = self.get_bucket(bucket_name)
-        path = '/'.join(components)
-
-        # the next few lines are equiv to `bucket.get_blob(path)`
-        # but without requiring storage.objects.get permission
-        blobs = list(self.client.list_blobs(bucket, prefix=path))
-        # first where r.name == path (or None)
-        return next((r for r in blobs if r.name == path), None)
-
-    def list_directory(self, directory_name) -> List[str]:
-        """List directory"""
-        path = self.file_path(directory_name)
-        if path.startswith('gs://'):
-            bucket_name, *components = directory_name[5:].split('/')
-            assert self.client
-            blobs = self.client.list_blobs(
-                bucket_name, prefix='/'.join(components), delimiter='/'
-            )
-            return [f'gs://{bucket_name}/{blob.name}' for blob in blobs]
-
-        return [os.path.join(path, f) for f in os.listdir(path)]
-
-    async def file_contents(self, filename) -> Optional[str]:
-        """Get contents of file (decoded as utf8)"""
-        path = self.file_path(filename)
-        if path.startswith('gs://'):
-            blob = await self.get_blob(path)
-            try:
-                retval = blob.download_as_bytes()
-                if isinstance(retval, bytes):
-                    retval = retval.decode()
-                return retval
-            except Forbidden:
-                logger.warning(f"FORBIDDEN: Can't download {filename}")
-                return None
-
-        with open(filename, encoding='utf-8') as f:
-            return f.read()
-
-    async def file_exists(self, filename: str) -> bool:
-        """Determines whether a file exists"""
-        path = self.file_path(filename)
-
-        if path.startswith('gs://'):
-            blob = await self.get_blob(filename)
-            return blob is not None
-
-        return os.path.exists(path)
-
-    async def file_size(self, filename):
-        """Get size of file in bytes"""
-        path = self.file_path(filename)
-        if path.startswith('gs://'):
-            blob = await self.get_blob(filename)
-            return blob.size
-
-        return os.path.getsize(path)
-
     @abstractmethod
     def get_sample_id(self, row: SingleRow) -> Optional[str]:
         """Get external sample ID from row"""
 
-    # @abstractmethod
-    async def get_cpg_sample_id(self, row: SingleRow) -> Optional[str]:
-        """Get internal cpg sample ID from row"""
-
     @abstractmethod
     def get_participant_id(self, row: SingleRow) -> Optional[str]:
         """Get external participant ID from row"""
@@ -373,7 +360,7 @@ class GenericParser:  # pylint: disable=too-many-public-methods
     # @abstractmethod
     def get_sequence_types(self, row: GroupedRow) -> List[SequenceType]:
         """Get sequence types from row"""
-        return List(SequenceType(self.default_sequence_type))
+        return [SequenceType(self.default_sequence_type)]
 
     # @abstractmethod
     def get_sequence_type(self, row: SingleRow) -> SequenceType:
@@ -400,6 +387,7 @@ class GenericParser:  # pylint: disable=too-many-public-methods
         rows: GroupedRow,
         external_sample_id: str,
         cpg_sample_id: Optional[str],
+        sequence_ids: Dict[Any, Any],
     ):
         """
         ASYNC function that (maps) transforms one GroupedRow, and returns a Tuple of:
@@ -412,13 +400,6 @@ class GenericParser:  # pylint: disable=too-many-public-methods
         Then the calling function does the (reduce).
         """
 
-        # Get all the sequence ids for this sample
-        sequence_ids = {}
-        if cpg_sample_id is not None:
-            sequence_ids = await self.seqapi.get_all_sequences_for_sample_id_async(
-                sample_id=cpg_sample_id
-            )
-
         # now we have sample / sequencing meta across 4 different rows, so collapse them
         (
             collapsed_sequencing_meta,
@@ -446,29 +427,28 @@ class GenericParser:  # pylint: disable=too-many-public-methods
                             f'Unhandled case with more than one sequence ID for the type {seq.sequence_type}'
                         )
                     sequence_id = sequence_id[0]
-                args = {
-                    'id': sequence_id,
+
+                seq_args: Dict[str, Any] = {
                     'meta': seq.meta,
                     'type': seq.sequence_type,
                     'status': self.get_sequence_status(seq.rows),
                 }
-                if not args['id']:
-                    del args['id']
+                if sequence_id:
+                    seq_args['id'] = sequence_id
 
-                sequences_to_upsert.append(SequenceUpsert(**args))
+                sequences_to_upsert.append(SequenceUpsert(**seq_args))
 
         # Should we add or update sample
-        args = {
-            'id': cpg_sample_id,
+        sample_args: Dict[str, Any] = {
             'meta': collapsed_sample_meta.meta,
             'external_id': external_sample_id,
             'type': self.get_sample_type(rows),
             'sequences': sequences_to_upsert,
         }
-        if not cpg_sample_id:
-            del args['id']
+        if cpg_sample_id:
+            sample_args['id'] = cpg_sample_id
 
-        sample_to_upsert = SampleBatchUpsert(**args)
+        sample_to_upsert = SampleBatchUpsert(**sample_args)
 
         if collapsed_analyses:
             analyses_to_add.extend(collapsed_analyses)
@@ -490,7 +470,12 @@ class GenericParser:  # pylint: disable=too-many-public-methods
         )
 
     async def process_participant_group(
-        self, participant_name: str, sample_map: Dict[str, Any]
+        self,
+        participant_name: str,
+        sample_map,
+        external_to_internal_sample_id_map: Dict[str, Any],
+        internal_pid: Optional[int],
+        sequence_map: Dict[str, Dict[str, int]] = None,
     ):
         """
         ASYNC function that (maps) transforms one GroupedRow, and returns a Tuple of:
@@ -506,36 +491,23 @@ class GenericParser:  # pylint: disable=too-many-public-methods
 
         all_rows = [r for row in sample_map.values() for r in row]
 
-        # Get external sid to cpg map
-        existing_external_id_to_cpgid = (
-            await SampleApi().get_sample_id_map_by_external_async(
-                self.sample_metadata_project,
-                list(sample_map.keys()),
-                allow_missing=True,
-            )
-        )
-
         # Get all the samples and sequences to upsert first
         samples_to_upsert = []
         sequences_to_upsert = []
         analyses_to_add = []
 
         for sample_id, rows in sample_map.items():
-            cpg_id = existing_external_id_to_cpgid.get(sample_id, None)
+            cpg_id = external_to_internal_sample_id_map.get(sample_id, None)
             sample, seqs, analyses = await self.process_sample_group(
-                rows, sample_id, cpg_id
+                rows=rows,
+                external_sample_id=sample_id,
+                cpg_sample_id=cpg_id,
+                sequence_ids=sequence_map.get(cpg_id, {}),
             )
             samples_to_upsert.append(sample)
             sequences_to_upsert.extend(seqs)
             analyses_to_add.extend(analyses)
 
-        # Construct participant to upsert
-        existing_participant_ids = self.papi.get_participant_id_map_by_external_ids(
-            self.sample_metadata_project, [participant_name], allow_missing=True
-        )
-
-        internal_id = existing_participant_ids.get(participant_name, None)
-
         # pull relevant participant fields
         reported_sex = self.get_reported_sex(all_rows)
         reported_gender = self.get_reported_gender(all_rows)
@@ -543,18 +515,16 @@ class GenericParser:  # pylint: disable=too-many-public-methods
 
         # now we have sample / sequencing meta across 4 different rows, so collapse them
         collapsed_participant_meta = await self.get_participant_meta(
-            internal_id, all_rows
+            internal_pid, all_rows
         )
 
-        args = {
-            'id': internal_id,  # noqa: E501
+        args: Dict[str, Any] = {
             'external_id': participant_name,
-            # 'reported_sex': None,
-            # 'reported_gender': None,
-            # 'karyotype': None,
             'meta': collapsed_participant_meta.meta,
             'samples': samples_to_upsert,
         }
+        if internal_pid:
+            args['id'] = internal_pid
 
         if reported_sex:
             args['reported_sex'] = reported_sex
@@ -563,9 +533,6 @@ class GenericParser:  # pylint: disable=too-many-public-methods
         if karyotype:
             args['karyotype'] = karyotype
 
-        if not internal_id:
-            del args['id']
-
         participant_to_upsert = ParticipantUpsert(**args)
 
         return (
@@ -579,7 +546,7 @@ class GenericParser:  # pylint: disable=too-many-public-methods
         self,
         file_pointer,
         delimiter: str,
-    ) -> Dict[str, List]:
+    ) -> Dict[str, List[Dict[str, Any]]]:
         """
         Parse manifest file into a list of dicts, indexed by sample name.
         """
@@ -596,40 +563,59 @@ class GenericParser:  # pylint: disable=too-many-public-methods
         Override this method if you can't use the default implementation that simply
         calls csv.DictReader
         """
-        reader = csv.DictReader(file_pointer, delimiter=delimiter)
+        reader = CustomDictReader(
+            file_pointer,
+            delimiter=delimiter,
+            key_map=self.key_map,
+            required_keys=self.required_keys,
+            ignore_extra_keys=self.ignore_extra_keys,
+        )
         return reader
 
     async def file_pointer_to_participant_map(
         self,
         file_pointer,
         delimiter: str,
-    ) -> Dict[Any, Dict[Any, List[Any]]]:
+    ) -> Dict[Any, Dict[str, List[Dict[str, Any]]]]:
         """
         Parse manifest file into a list of dicts, indexed by participant id.
         """
 
-        participant_map: Dict[Any, Dict[Any, List[Any]]] = defaultdict(
+        participant_map: Dict[str, Dict[Any, List[Any]]] = defaultdict(
             lambda: defaultdict(list)
         )
         reader = self._get_dict_reader(file_pointer, delimiter=delimiter)
         for row in reader:
-            pid = self.get_participant_id(row)
+            participant_eid = self.get_participant_id(row)
 
-            if not pid:
+            if not participant_eid:
                 raise ValueError(f'Participant not found in row: {row}')
 
-            sid = self.get_sample_id(row)
-            participant_map[pid][sid].append(row)
+            # business rule to default sample ID to participant ID if not provided
+            sid = self.get_sample_id(row) or participant_eid
+            participant_map[participant_eid][sid].append(row)
 
         return participant_map
 
-    async def validate_rows(self, sample_map: Dict[str, Union[Dict, List[Dict]]]):
+    async def validate_participant_map(
+        self, participant_map: Dict[Any, Dict[str, List[Dict[str, Any]]]]
+    ):
         """
         Validate sample rows:
         - throw an exception if an error occurs
         - log a warning for all other issues
         """
-        return
+        if len(participant_map) == 0:
+            raise ValueError('The manifest contains no records')
+
+    async def validate_sample_map(self, sample_map: Dict[str, List[Dict[str, Any]]]):
+        """
+        Validate sample rows:
+        - throw an exception if an error occurs
+        - log a warning for all other issues
+        """
+        if len(sample_map) == 0:
+            raise ValueError('The manifest contains no records')
 
     async def parse_manifest(  # pylint: disable=too-many-branches
         self, file_pointer, delimiter=',', confirm=False, dry_run=False
@@ -639,27 +625,20 @@ class GenericParser:  # pylint: disable=too-many-public-methods
 
         Returns a dict mapping external sample ID to CPG sample ID
         """
-        sample_map: Dict[str, Any] = {}
-        participant_map: Dict[str, Any] = {}
         if self.has_participants(file_pointer, delimiter):
             participant_map = await self.file_pointer_to_participant_map(
                 file_pointer, delimiter
             )
-        else:
-            sample_map = await self.file_pointer_to_sample_map(file_pointer, delimiter)
-
-        if len(sample_map) == 0 and len(participant_map) == 0:
-            raise ValueError(
-                f'{self.sample_metadata_project}: The manifest file contains no records'
+            await self.validate_participant_map(participant_map)
+            return await self.parse_manifest_by_participants(
+                participant_map, confirm=confirm, dry_run=dry_run
             )
 
-        if len(sample_map) != 0:
-            return await self.parse_manifest_by_samples(
-                sample_map, confirm=confirm, dry_run=dry_run
-            )
+        sample_map = await self.file_pointer_to_sample_map(file_pointer, delimiter)
+        await self.validate_sample_map(sample_map)
 
-        return await self.parse_manifest_by_participants(
-            participant_map, confirm=confirm, dry_run=dry_run
+        return await self.parse_manifest_by_samples(
+            sample_map, confirm=confirm, dry_run=dry_run
         )
 
     def upsert_summary(
@@ -698,7 +677,7 @@ class GenericParser:  # pylint: disable=too-many-public-methods
 
     async def add_analyses(self, analyses_to_add, external_to_internal_id_map):
         """Given an analyses dictionary add analyses"""
-        proj = self.sample_metadata_project
+        proj = self.project
         analysisapi = AnalysisApi()
 
         logger.info(
@@ -732,7 +711,7 @@ class GenericParser:  # pylint: disable=too-many-public-methods
     ) -> Dict[str, Any]:
         """Parses a manifest of data that is keyed on participant id and sample id"""
 
-        proj = self.sample_metadata_project
+        proj = self.project
 
         # all dicts indexed by external_sample_id
         summary: Dict[str, Dict[str, List[Any]]] = None
@@ -746,9 +725,43 @@ class GenericParser:  # pylint: disable=too-many-public-methods
             if self.verbose:
                 logger.info(f'{proj}:Preparing {", ".join(external_pids)}')
 
+            # Construct participant to upsert
+            existing_participant_ids = (
+                await self.papi.get_participant_id_map_by_external_ids_async(
+                    self.project, external_pids, allow_missing=True
+                )
+            )
+
+            sample_ids = list(
+                set(k for s in participant_map.values() for k in s.keys())
+            )
+
+            external_to_internal_sample_id_map = (
+                await self.sapi.get_sample_id_map_by_external_async(
+                    self.project, sample_ids, allow_missing=True
+                )
+            )
+
+            sequences = []
+            if external_to_internal_sample_id_map:
+                sequences = await self.seqapi.get_sequences_by_sample_ids_async(
+                    list(external_to_internal_sample_id_map.values()),
+                    get_latest_sequence_only=False,
+                )
+
+            sequence_map: Dict[str, Dict[str, int]] = defaultdict(dict)
+            for seq in sequences:
+                sequence_map[seq['sample_id']][seq['type']] = seq['id']
+
             for external_pid in external_pids:
                 sample_map = participant_map[external_pid]
-                promise = self.process_participant_group(external_pid, sample_map)
+                promise = self.process_participant_group(
+                    participant_name=external_pid,
+                    sample_map=sample_map,
+                    internal_pid=existing_participant_ids.get(external_pid),
+                    sequence_map=sequence_map,
+                    external_to_internal_sample_id_map=external_to_internal_sample_id_map,
+                )
                 current_batch_promises[external_pid] = promise
 
             processed_ex_pids = list(current_batch_promises.keys())
@@ -780,17 +793,24 @@ class GenericParser:  # pylint: disable=too-many-public-methods
                     {external_pid: analysis_to_add},
                 )
 
+        sequences_count: Dict[str, int] = defaultdict(int)
+        for s in summary['sequences']['insert'] + summary['sequences']['update']:
+            sequences_count[str(s['type'])] += 1
+
+        str_seq_count = ', '.join(f'{k}={v}' for k, v in sequences_count.items())
         message = f"""\
-            {proj}: Processing participants: {', '.join(participant_map.keys())}
+    {proj}: Processing participants: {', '.join(participant_map.keys())}
 
-            Adding {len(summary['participants']['insert'])} participants
-            Adding {len(summary['samples']['insert'])} samples
-            Adding {len(summary['sequences']['insert'])} sequences
-            Adding {len(summary['analyses']['insert'])} analysis
+    Sequence types: {str_seq_count}
 
-            Updating {len(summary['participants']['update'])} participants
-            Updating {len(summary['samples']['update'])} samples
-            Updating {len(summary['sequences']['update'])} sequences
+    Adding {len(summary['participants']['insert'])} participants
+    Adding {len(summary['samples']['insert'])} samples
+    Adding {len(summary['sequences']['insert'])} sequences
+    Adding {len(summary['analyses']['insert'])} analysis
+
+    Updating {len(summary['participants']['update'])} participants
+    Updating {len(summary['samples']['update'])} samples
+    Updating {len(summary['sequences']['update'])} sequences
         """
 
         if dry_run:
@@ -815,13 +835,12 @@ class GenericParser:  # pylint: disable=too-many-public-methods
         self, sample_map: Dict[str, Any], confirm: bool = False, dry_run: bool = False
     ) -> Dict[str, Any]:
         """Parses a manifest of data that is keyed on sample id"""
-        proj = self.sample_metadata_project
+        proj = self.project
 
         # now we can start adding!!
-        sapi = SampleApi()
 
         # Map external sids into cpg ids
-        existing_external_id_to_cpgid = await sapi.get_sample_id_map_by_external_async(
+        existing_external_id_to_cpgid = await self.sapi.get_sample_id_map_by_external_async(
             proj, list(sample_map.keys()), allow_missing=True
         )
 
@@ -837,10 +856,19 @@ class GenericParser:  # pylint: disable=too-many-public-methods
             if self.verbose:
                 logger.info(f'{proj}:Preparing {", ".join(external_sids)}')
 
+            sequence_ids = await self.seqapi.get_sequence_ids_from_sample_ids_async(
+                external_sids
+            )
+
             for external_sid in external_sids:
                 rows: GroupedRow = sample_map[external_sid]
                 cpg_id = existing_external_id_to_cpgid.get(external_sid, None)
-                promise = self.process_sample_group(rows, external_sid, cpg_id)
+                promise = self.process_sample_group(
+                    rows=rows,
+                    external_sample_id=external_sid,
+                    cpg_sample_id=cpg_id,
+                    sequence_ids=sequence_ids,
+                )
                 current_batch_promises[external_sid] = promise
 
             processed_ex_sids = list(current_batch_promises.keys())
@@ -895,13 +923,13 @@ class GenericParser:  # pylint: disable=too-many-public-methods
             logger.info(message)
 
         # Batch update
-        result = sapi.batch_upsert_samples(
+        result = await self.sapi.batch_upsert_samples_async(
             proj, SampleBatchUpsertBody(samples=all_samples)
         )
 
         # Add analyses
         # Map external sids into cpg ids
-        existing_external_id_to_cpgid = await sapi.get_sample_id_map_by_external_async(
+        existing_external_id_to_cpgid = await self.sapi.get_sample_id_map_by_external_async(
             proj, list(sample_map.keys()), allow_missing=True
         )
         _ = await self.add_analyses(analyses_to_add, existing_external_id_to_cpgid)
@@ -909,7 +937,7 @@ class GenericParser:  # pylint: disable=too-many-public-methods
         return result
 
     async def parse_files(
-        self, sample_id: str, reads: List[str]
+        self, sample_id: str, reads: List[str], checksums: List[str] = None
     ) -> Dict[SUPPORTED_FILE_TYPE, Dict[str, List]]:
         """
         Returns a tuple of:
@@ -917,6 +945,19 @@ class GenericParser:  # pylint: disable=too-many-public-methods
         2. parsed type (fastq, cram, bam)
         """
 
+        if not isinstance(reads, list):
+            reads = [reads]
+
+        if not checksums:
+            checksums = [None] * len(reads)
+
+        if len(checksums) != len(reads):
+            raise ValueError(
+                'Expected length of reads to match length of provided checksums'
+            )
+
+        read_to_checksum = dict(zip(reads, checksums))
+
         file_by_type: Dict[SUPPORTED_FILE_TYPE, Dict[str, List]] = defaultdict(
             lambda: defaultdict(list)
         )
@@ -929,7 +970,8 @@ class GenericParser:  # pylint: disable=too-many-public-methods
             fastq_files: List[Sequence[Union[Coroutine, BaseException]]] = []  # type: ignore
             for fastq_group in structured_fastqs:
                 create_file_futures: List[Coroutine] = [
-                    self.create_file_object(f) for f in fastq_group
+                    self.create_file_object(f, checksum=read_to_checksum.get(f))
+                    for f in fastq_group
                 ]
                 fastq_files.append(asyncio.gather(*create_file_futures))  # type: ignore
 
@@ -1033,6 +1075,17 @@ class GenericParser:  # pylint: disable=too-many-public-methods
         >>> GenericParser.parse_fastqs_structure(['/directory1/Z01_1234_HNXXXXX_TCATCCTT-AGCGAGCT_L001_R1.fastq.gz', '/directory2/Z01_1234_HNXXXXX_TCATCCTT-AGCGAGCT_L001_R2.fastq.gz'])
         [['/directory1/Z01_1234_HNXXXXX_TCATCCTT-AGCGAGCT_L001_R1.fastq.gz', '/directory2/Z01_1234_HNXXXXX_TCATCCTT-AGCGAGCT_L001_R2.fastq.gz']]
 
+        >>> GenericParser.parse_fastqs_structure(['Sample_1_L01_1.fastq.gz', 'Sample_1_L01_2.fastq.gz', 'Sample_1_L02_R1.fastq.gz', 'Sample_1_L02_R2.fastq.gz'])
+        [['Sample_1_L01_1.fastq.gz', 'Sample_1_L01_2.fastq.gz'], ['Sample_1_L02_R1.fastq.gz', 'Sample_1_L02_R2.fastq.gz']]
+
+        >>> GenericParser.parse_fastqs_structure(['File_1.fastq', 'File_2.fastq'])
+        [['File_1.fastq', 'File_2.fastq']]
+
+        >>> GenericParser.parse_fastqs_structure(['File_1.fq', 'File_2.fq'])
+        [['File_1.fq', 'File_2.fq']]
+
+        >>> GenericParser.parse_fastqs_structure(['File_1.fq.gz', 'File_2.fq.gz'])
+        [['File_1.fq.gz', 'File_2.fq.gz']]
         """
         # find last instance of R\d, and then group by prefix on that
         sorted_fastqs = sorted(fastqs)
@@ -1053,6 +1106,9 @@ class GenericParser:  # pylint: disable=too-many-public-methods
             sorted_fastqs, lambda r: r_matches[r][0][: r_matches[r][1].start()]  # type: ignore
         ):
             values.append(sorted(grouped))
+        invalid_fastq_groups = [grp for grp in values if len(grp) != 2]
+        if invalid_fastq_groups:
+            raise ValueError(f'Invalid fastq group {invalid_fastq_groups}')
 
         return sorted(values, key=lambda el: el[0])
 
@@ -1060,17 +1116,19 @@ class GenericParser:  # pylint: disable=too-many-public-methods
         self,
         filename: str,
         secondary_files: List[SingleRow] = None,
+        checksum: Optional[str] = None,
     ) -> SingleRow:
         """Takes filename, returns formed CWL dictionary"""
-        checksum = None
+        _checksum = checksum
         file_size = None
 
         if not self.skip_checking_gcs_objects:
-            md5_filename = self.file_path(filename + '.md5')
-            if await self.file_exists(md5_filename):
-                contents = await self.file_contents(md5_filename)
-                if contents:
-                    checksum = f'md5:{contents.strip()}'
+            if not _checksum:
+                md5_filename = self.file_path(filename + '.md5')
+                if await self.file_exists(md5_filename):
+                    contents = await self.file_contents(md5_filename)
+                    if contents:
+                        _checksum = f'md5:{contents.strip()}'
 
             file_size = await self.file_size(filename)
 
@@ -1078,7 +1136,7 @@ class GenericParser:  # pylint: disable=too-many-public-methods
             'location': self.file_path(filename),
             'basename': os.path.basename(filename),
             'class': 'File',
-            'checksum': checksum,
+            'checksum': _checksum,
             'size': file_size,
         }
 
@@ -1121,6 +1179,16 @@ class GenericParser:  # pylint: disable=too-many-public-methods
         if relevant_delimiter:
             return relevant_delimiter
 
+        # pylint: disable=no-member
+        with AnyPath(filename).open('r') as f:
+            first_line = f.readline()
+            delimiter = csv.Sniffer().sniff(first_line).delimiter
+            if delimiter:
+                logger.info(
+                    f'Guessing delimiter based on first line, got "{delimiter}"'
+                )
+                return delimiter
+
         raise ValueError(f'Unrecognised extension on file: {filename}')
 
 
